# $Date: 1997/12/05 08:46:51 $  $Author: kg $  $Revision: 1.30 $ #
# kg, 10/08/93 #

#++
D -- differential operator

D(f)
D([i1,...,in], f)
D([],f)

f - expression
[i1,...,in] - list of positive integers

D(f) computes the derivative of the function f. D([i],f) computes the
partial derivative of f with respect to its i-th argument. D([i,j],f)
is equivalent to D([i],D([j],f)).
++#

D:= proc()
    local idx, f, x, mapD, i, n, nidx;
begin
    if args(0) = 1 then
	idx:= null(); nidx:= 1; f:= args(1); mapD:= D;
    elif args(0) = 2 then
	idx:= args(1);
	if domtype(idx) <> DOM_LIST then
	    error("invalid indices")
	end_if;
	nidx:= nops(idx);
	f:= args(2);
	mapD:= fun(D(args(2), args(1)));
    else error("wrong no of args") end_if;

    if nidx = 0 then return(f) end_if;
    if testargs() then
	if domtype(idx) = DOM_LIST then
	    if not testtype(idx, Type::ListOf(Type::PosInt)) then
		error("indices must be positive integers")
	    end_if
	end_if
    end_if;

    case domtype(f)
    of DOM_LIST do
    of DOM_SET do
    of DOM_TABLE do
    of DOM_ARRAY do
	return(map(f, mapD, idx))

    of DOM_EXPR do
	break;

    of DOM_POLY do
	return(Dpoly(idx, f));

    of DOM_PROC do
	x:= [op(f,1)];
	if domtype(idx) = DOM_LIST then
	    if max(op(idx)) > nops(x) then
		error("procedure hasn't that much unknowns")
	    end_if;
	    return(stdlib::makeDop(f, x, idx))
	end_if;
	if nops(x) <> 1 then
	    error("procedure has more than one argument")
	end_if;
        if type(op(f,4)) <> "_if" and
           traperror((n:=stdlib::makeDop(f, x)))=0 then return(n)
        # stdlib::procdiff deals the case where the procedure
          cannot be called with symbolic parameters #
        else return(stdlib::mapproc(f,diff,op(x)))
        end_if;

    of DOM_EXEC do
    	if domtype(idx) <> DOM_LIST then
	    return(stdlib::makeDop(f, [genident()]))
	end_if;
	return(procname(idx, f));

    of DOM_FUNC_ENV do
	if funcattr(f, "D") <> FAIL then
	    return(funcattr(f, "D")(idx, f))
	end_if;
	if contains({DOM_PROC, DOM_EXEC}, domtype(op(f,1))) then
	    return(D(idx, op(f,1)))
	end_if;
	if domtype(idx) <> DOM_LIST then
	    return(stdlib::makeDop(f, [genident()]))
	end_if;
	return(procname(idx, f));

    of DOM_IDENT do
	if testtype(f, Type::Constant) then return(0) end_if;
	return(procname(idx, f));

    otherwise
	return((if f::D <> FAIL then f::D else procname end_if)(idx, f));
    end_case;

    if nidx = 1 then
	case type(f)
	of "_negate" do
	    return(-D(idx, op(f,1)));

	of "_plus" do
	    return(map(f, mapD, idx));

	of "_subtract" do
	    return(D(idx, op(f,1) - op(f,2)));

	of "_mult" do
	    return(_plus(subsop(f, i=D(idx,op(f,i))) $ i=1..nops(f)));

	of "_divide" do
	    return(D(idx, op(f,1) / op(f,2)));

	of "_invert" do
	    return(- op(f,1)^(-2) * D(idx, op(f,1)));

	of "_power" do
	    if testtype(op(f,2), NUMERIC) then
		return(op(f,2)*D(idx, op(f,1))*op(f,1)^(op(f,2)-1));
	    end_if;
	    return(procname(idx, f));

	of "_fconcat" do
	    n:= nops(f);
	    return(_mult(D(idx, op(f,n)),
			 (_fconcat(D(op(f,i)), op(f, i+1..n))
			  $ i=1..n-1)));

	of "_index" do
	    return(procname(idx, f));
	end_case;
    else
	case type(f)
	of "_negate" do
	    return(-D(idx, op(f,1)));

	of "_plus" do
	    return(map(f, mapD, idx));

	of "_subtract" do
	    return(D(idx, op(f,1) - op(f,2)));

	of "_mult" do
	    x:= idx[1];
	    idx[1]:= NIL;
	    return(D([x], _plus(subsop(f, i=D(idx,op(f,i))) $ i=1..nops(f))));

	of "_divide" do
	    return(D(idx, op(f,1) / op(f,2)));

	of "_invert" do
	    x:= idx[1];
	    idx[1]:= NIL;
	    return(D([x], - op(f,1)^(-2) * D(idx, op(f,1))));

	of "_power" do
	    if testtype(op(f,2), NUMERIC) then
		x:= idx[1];
		idx[1]:= NIL;
		return(D([x], op(f,2)*D(idx, op(f,1))*op(f,1)^(op(f,2)-1)));
	    end_if;
	    return(procname(idx, f));

	of "_fconcat" do
	    n:= nops(f);
	    x:= idx[1];
	    idx[1]:= NIL;
	    return(D([x], _mult(D(idx, op(f,n)),
			  (_fconcat(D(op(f,i)), op(f, i+1..n))
			  $ i=1..n-1))));

	of "_index" do
	    return(procname(idx, f));
	end_case;
    end_if;

    if domtype(idx) = DOM_LIST then
	if type(f) = "D" then
	    if nops(f) = 2 then
		if domtype(op(f,1)) = DOM_LIST then
		    return(procname(idx.op(f,1), op(f,2)))
		end_if
	    end_if
	end_if
    end_if;

    if testtype(f, Type::Constant) then return(0) end_if;
    procname(idx, f)
end_proc:

D:= funcattr(D, "type", "D"):
D:= funcattr(D, "print", "D"):

#--
stdlib::makeDop -- reconstruct differential operator with diff

stdlib::makeDop(f, x [, idx])

f   - function for which to create D operator
x   - list of unknowns of f
idx - list of partial derivatives, [1] is default

Reconstructs D(idx,f) by creating a function from diff(f(op(x)), x[idx[1]]...)
--#

stdlib::makeDop:= proc(f, x, idx)
    local v, dv, i;
begin
    v:= f(op(x));
    if args(0) = 3 then
        dv:= diff(v, x[idx[i]] $ hold(i)=1..nops(idx))
    else
    	dv:= diff(v, x[1])
    end_if;
    if type(dv) = "diff" then
        hold(D)((if args(0) = 3 then idx else null() end_if), op(v,0))
    else
        v:= fp::unapply(dv, op(x));
        if v = FAIL then
	    newpurefunc(expr2text(hold(func)(dv, op(x))), NIL,
		misc::subsFreeVars(dv, (x[i]=newfuncarg(i)) $ hold(i)=1..nops(x),
				   Unsimplified))
        else
            v
        end_if
    end_if
end_proc:

# stdlib::procdiff(f,[u,v,...]) computes the derivative of
  the procedure f with respect to u,v,... without evaluating it,
  for example
>> a:=proc(x) begin if x < 0 then -x else x end_if end_proc:
>> stdlib::procdiff(a,[x]);
                             proc(x)
                               name a;
                             begin
                               if x < 0 then
                                 -1
                               else
                                 1
                               end_if
                             end_proc
#
stdlib::mapproc := proc(f)
begin
   case type(f)
   of DOM_PROC do
      if type(op(f,4))="_stmtseq" then FAIL
      else subsop(f,4=stdlib::mapproc(op(f,4),args(2..args(0))),6=NIL)
      end_if; break
   of "_if" do
      subsop(f,2=stdlib::mapproc(op(f,2),args(2..args(0))),
               3=stdlib::mapproc(op(f,3),args(2..args(0)))); break
   of "_stmtseq" do FAIL
   of "function" do
      if op(f,0)=hold(procname) then return(FAIL) end_if
   otherwise args(2)(f,args(3..args(0)))
   end_case
end_proc:

# end of file #
